/*
 * ypauth: YP authentication DLL for Windows 2000
 * Copyright (C) Matthew Chapman 2001-2002
 */

#include "ypauth.h"

/* LSA function table */
PLSA_DISPATCH_TABLE			LSA;

/* Handle to real MSV1_0 DLL */
static HINSTANCE			MSV1_0_DLL;

/* Information about the local domain */
static POLICY_ACCOUNT_DOMAIN_INFO *	DomainInfo;
static UNICODE_STRING			Administrator;

/* MSV1_0 Authentication package functions */
static PLSA_AP_INITIALIZE_PACKAGE       MSV1_0_InitializePackage;
static PLSA_AP_LOGON_USER_EX2           MSV1_0_LogonUserEx2;
static PLSA_AP_LOGON_TERMINATED         MSV1_0_LogonTerminated;
static PLSA_AP_CALL_PACKAGE             MSV1_0_CallPackage;
static PLSA_AP_CALL_PACKAGE_PASSTHROUGH MSV1_0_CallPackagePassthrough;
static PLSA_AP_CALL_PACKAGE_UNTRUSTED   MSV1_0_CallPackageUntrusted;
static PLSA_AP_MS_INITIALIZE            MSV1_0_MsInitialize;

/* MSV1_0 SSPI functions */
static SpLsaModeInitializeFn		MSV1_0_SpLsaModeInitialize;
static SpUserModeInitializeFn		MSV1_0_SpUserModeInitialize;

/* Non API functions exported from MSV1_0 */
#ifdef EXPORT_MSV_INTERFACE
static MSV_SAM_VALIDATE			MSV1_0_SamValidate;
static MSV_SAM_LOGOFF			MSV1_0_SamLogoff;
static MSV_GET_LOGON_ATTEMPT_COUNT	MSV1_0_GetLogonAttemptCount;
static MSV_EXPORT_SUB_AUTHENTICATION_ROUTINE	MSV1_0_ExportSubAuthenticationRoutine;
static MSV_SUB_AUTHENTICATION_PRESENT	MSV1_0_SubAuthenticationPresent;
#endif


BOOL WINAPI DllMain(
    HINSTANCE hinstDLL,
    DWORD fdwReason,
    LPVOID lpvReserved
  )
{
	switch (fdwReason)
	{
	    case DLL_PROCESS_ATTACH:
		rpc_nt_init();

		/* Load real library and initialise thunks */
		MSV1_0_DLL = LoadLibrary(MSV1_0_DLL_NAME);

		MSV1_0_SpLsaModeInitialize = (SpLsaModeInitializeFn)
			GetProcAddress(MSV1_0_DLL, SECPKG_LSAMODEINIT_NAME);
		MSV1_0_SpUserModeInitialize = (SpUserModeInitializeFn)
			GetProcAddress(MSV1_0_DLL, SECPKG_USERMODEINIT_NAME);

		MSV1_0_InitializePackage = (PLSA_AP_INITIALIZE_PACKAGE)
			GetProcAddress(MSV1_0_DLL, LSA_AP_NAME_INITIALIZE_PACKAGE);
		MSV1_0_LogonUserEx2 = (PLSA_AP_LOGON_USER_EX2)
			GetProcAddress(MSV1_0_DLL, LSA_AP_NAME_LOGON_USER_EX2);
		MSV1_0_LogonTerminated = (PLSA_AP_LOGON_TERMINATED)
			GetProcAddress(MSV1_0_DLL, LSA_AP_NAME_LOGON_TERMINATED);
		MSV1_0_CallPackage = (PLSA_AP_CALL_PACKAGE)
			GetProcAddress(MSV1_0_DLL, LSA_AP_NAME_CALL_PACKAGE);
		MSV1_0_CallPackagePassthrough = (PLSA_AP_CALL_PACKAGE_PASSTHROUGH)
			GetProcAddress(MSV1_0_DLL, LSA_AP_NAME_CALL_PACKAGE_PASSTHROUGH);
		MSV1_0_CallPackageUntrusted = (PLSA_AP_CALL_PACKAGE_UNTRUSTED)
			GetProcAddress(MSV1_0_DLL, LSA_AP_NAME_CALL_PACKAGE_UNTRUSTED);
		MSV1_0_MsInitialize = (PLSA_AP_MS_INITIALIZE)
			GetProcAddress(MSV1_0_DLL, LSA_AP_NAME_MS_INITIALIZE);

#ifdef EXPORT_MSV_INTERFACE
		MSV1_0_SamValidate = (MSV_SAM_VALIDATE)
			GetProcAddress(MSV1_0_DLL, "MsvSamValidate");
		MSV1_0_SamLogoff = (MSV_SAM_LOGOFF)
			GetProcAddress(MSV1_0_DLL, "MsvSamLogoff");
		MSV1_0_GetLogonAttemptCount = (MSV_GET_LOGON_ATTEMPT_COUNT)
			GetProcAddress(MSV1_0_DLL, "MsvGetLogonAttemptCount");
		MSV1_0_ExportSubAuthenticationRoutine = (MSV_EXPORT_SUB_AUTHENTICATION_ROUTINE)
			GetProcAddress(MSV1_0_DLL, "Msv1_0ExportSubAuthenticationRoutine");
		MSV1_0_SubAuthenticationPresent = (MSV_SUB_AUTHENTICATION_PRESENT)
			GetProcAddress(MSV1_0_DLL, "Msv1_0SubAuthenticationPresent");
#endif
		break;

	    case DLL_PROCESS_DETACH:
		FreeLibrary(MSV1_0_DLL);
		rpc_nt_exit();
		break;
	}

	return TRUE;
}


/*
 * LSA Authentication Package interface
 */

NTSTATUS NTAPI
  LsaApInitializePackage(
    ULONG AuthenticationPackageId,
    PLSA_DISPATCH_TABLE LsaDispatchTable,
    PLSA_STRING Database OPTIONAL,
    PLSA_STRING Confidentiality OPTIONAL,
    PLSA_STRING* AuthenticationPackageName
  )
{
	LSA = LsaDispatchTable;

	return MSV1_0_InitializePackage(AuthenticationPackageId, LsaDispatchTable,
		Database, Confidentiality, AuthenticationPackageName);
}


#define REMAP_UNICODE_STRING(String) { \
	ULONG Offset = (PCHAR)(String.Buffer) - (PCHAR)(ClientBufferBase); \
	if ((Offset+String.Length) > SubmitBufferSize) \
		return STATUS_INVALID_PARAMETER; \
	String.Buffer = (LPWSTR)((PCHAR)(ProtocolSubmitBuffer) + Offset); \
}

NTSTATUS NTAPI
  LsaApLogonUserEx2(
     PLSA_CLIENT_REQUEST ClientRequest,
     SECURITY_LOGON_TYPE LogonType,
     PVOID ProtocolSubmitBuffer,
     PVOID ClientBufferBase,
     ULONG SubmitBufferSize,
     PVOID* ProfileBuffer,
     PULONG ProfileBufferSize,
     PLUID LogonId,
     PNTSTATUS SubStatus,
     PLSA_TOKEN_INFORMATION_TYPE TokenInformationType,
     PVOID* TokenInformation,
     PUNICODE_STRING* AccountName,
     PUNICODE_STRING* AuthenticatingAuthority,
     PUNICODE_STRING* MachineName,
     PSECPKG_PRIMARY_CRED PrimaryCredentials,
     PSECPKG_SUPPLEMENTAL_CRED_ARRAY* SupplementalCredentials
  )
{
	MSV1_0_INTERACTIVE_LOGON *Logon = (MSV1_0_INTERACTIVE_LOGON *)ProtocolSubmitBuffer;
	LSA_TOKEN_INFORMATION_V1 *Token;
	LPWSTR ClientUserName, ClientDomainName;
	STRING UserName, Password;
	UCHAR DecodeSeed;
	NTSTATUS Status;
	USHORT Uid;

	/* We do this here since the LSA RPC server is not ready at initialisation time */
	if (DomainInfo == NULL)
	{
		Status = InitializeDomain(&DomainInfo);
		if (Status != STATUS_SUCCESS)
			return Status;

		RtlInitUnicodeString(&Administrator, ADMINISTRATOR);
	}

	if (Logon->MessageType == MsV1_0InteractiveLogon)
	{
		ClientUserName = Logon->UserName.Buffer;
		ClientDomainName = Logon->LogonDomainName.Buffer;

		REMAP_UNICODE_STRING(Logon->UserName);
		REMAP_UNICODE_STRING(Logon->LogonDomainName);
		if ((RtlCompareUnicodeString(&Logon->LogonDomainName, &DomainInfo->DomainName, TRUE) == 0)
			&& (RtlCompareUnicodeString(&Logon->UserName, &Administrator, TRUE) != 0))
		{
			/* Attempt logon against YP */
			DecodeSeed = (UCHAR)(Logon->Password.Length >> 8);
			Logon->Password.Length &= 0xff;
			REMAP_UNICODE_STRING(Logon->Password);

			if (DecodeSeed != 0)
				RtlRunDecodeUnicodeString(DecodeSeed, &Logon->Password);

			RtlUnicodeStringToAnsiString(&Password, &Logon->Password, TRUE);
			RtlUnicodeStringToAnsiString(&UserName, &Logon->UserName, TRUE);

			Status = YpLogon(YP_SERVER_1, &UserName, &Password, &Uid);
			if (Status == STATUS_NO_LOGON_SERVERS)
				Status = YpLogon(YP_SERVER_2, &UserName, &Password, &Uid);

			RtlZeroMemory(Password.Buffer, Password.Length);
			RtlFreeAnsiString(&Password);
			RtlFreeAnsiString(&UserName);

			if (Status != STATUS_SUCCESS)
				return Status; /* failure */

			CreateProfile(ClientRequest, ProfileBuffer, ProfileBufferSize);

			*TokenInformationType = LsaTokenInformationV1;
			Token = LSA->AllocateLsaHeap(sizeof(LSA_TOKEN_INFORMATION_V1));
			RtlZeroMemory(Token, sizeof(LSA_TOKEN_INFORMATION_V1));
			Token->User.User.Sid = AppendSubAuthority(DomainInfo->DomainSid, YP_RID_BASE + Uid);
			Token->PrimaryGroup.PrimaryGroup = AppendSubAuthority(DomainInfo->DomainSid, DOMAIN_GROUP_RID_USERS);
			*TokenInformation = Token;

			AllocateLocallyUniqueId(LogonId);
			LSA->CreateLogonSession(LogonId);
			*AccountName = DuplicateUnicodeString(&Logon->UserName);
			*AuthenticatingAuthority = DuplicateUnicodeString(&Logon->LogonDomainName);
			*MachineName = DuplicateUnicodeString(&Logon->LogonDomainName);

			RtlZeroMemory(PrimaryCredentials, sizeof(SECPKG_PRIMARY_CRED));
			PrimaryCredentials->LogonId = *LogonId;
			CopyUnicodeString(&PrimaryCredentials->DownlevelName, &Logon->UserName);
			CopyUnicodeString(&PrimaryCredentials->DomainName, &Logon->LogonDomainName);
			CopyUnicodeString(&PrimaryCredentials->Password, &Logon->Password);
			PrimaryCredentials->UserSid = DuplicateSid(Token->User.User.Sid);
			PrimaryCredentials->Flags = PRIMARY_CRED_CLEAR_PASSWORD;
			CopyUnicodeString(&PrimaryCredentials->LogonServer, &Logon->LogonDomainName);
			*SupplementalCredentials = NULL;

			return STATUS_SUCCESS;
		}

		Logon->UserName.Buffer = ClientUserName;
		Logon->LogonDomainName.Buffer = ClientDomainName;
	}

	return MSV1_0_LogonUserEx2(ClientRequest, LogonType, ProtocolSubmitBuffer, ClientBufferBase,
		SubmitBufferSize, ProfileBuffer, ProfileBufferSize, LogonId, SubStatus, TokenInformationType,
		TokenInformation, AccountName, AuthenticatingAuthority, MachineName, PrimaryCredentials,
		SupplementalCredentials);
}


VOID NTAPI
  LsaApLogonTerminated(
    PLUID LogonId
  )
{
	MSV1_0_LogonTerminated(LogonId);
}


NTSTATUS NTAPI
  LsaApCallPackage(
    PLSA_CLIENT_REQUEST ClientRequest,
    PVOID ProtocolSubmitBuffer,
    PVOID ClientBufferBase,
    ULONG SubmitBufferLength,
    PVOID* ProtocolReturnBuffer,
    PULONG ReturnBufferLength,
    PNTSTATUS ProtocolStatus
  )
{
	return MSV1_0_CallPackage(ClientRequest, ProtocolSubmitBuffer, ClientBufferBase,
		SubmitBufferLength, ProtocolReturnBuffer, ReturnBufferLength, ProtocolStatus);
}


NTSTATUS NTAPI
  LsaApCallPackagePassthrough(
    PLSA_CLIENT_REQUEST ClientRequest,
    PVOID ProtocolSubmitBuffer,
    PVOID ClientBufferBase,
    ULONG SubmitBufferLength,
    PVOID* ProtocolReturnBuffer,
    PULONG ReturnBufferLength,
    PNTSTATUS ProtocolStatus
  )
{
	return MSV1_0_CallPackagePassthrough(ClientRequest, ProtocolSubmitBuffer, ClientBufferBase,
		SubmitBufferLength, ProtocolReturnBuffer, ReturnBufferLength, ProtocolStatus);
}


NTSTATUS NTAPI
  LsaApCallPackageUntrusted(
    PLSA_CLIENT_REQUEST ClientRequest,
    PVOID ProtocolSubmitBuffer,
    PVOID ClientBufferBase,
    ULONG SubmitBufferLength,
    PVOID* ProtocolReturnBuffer,
    PULONG ReturnBufferLength,
    PNTSTATUS ProtocolStatus
  )
{
	return MSV1_0_CallPackageUntrusted(ClientRequest, ProtocolSubmitBuffer, ClientBufferBase,
		SubmitBufferLength, ProtocolReturnBuffer, ReturnBufferLength, ProtocolStatus);
}


NTSTATUS NTAPI
  LsaApMsInitialize(
    PVOID Param
  )
{
	return MSV1_0_MsInitialize(Param);
}


/*
 * SSPI interface
 */

NTSTATUS NTAPI
  SpLsaModeInitialize(
    ULONG LsaVersion,
    PULONG PackageVersion,
    PSECPKG_FUNCTION_TABLE *ppTables,
    PULONG pcTables
  )
{
	NTSTATUS Status;

	Status = MSV1_0_SpLsaModeInitialize(LsaVersion, PackageVersion, ppTables, pcTables);
	if (Status != STATUS_SUCCESS)
		return Status;

	/* Hook LsaApLogonUserEx2 */
	(*ppTables)->LogonUserEx2 = LsaApLogonUserEx2;
	return STATUS_SUCCESS;
}


NTSTATUS NTAPI
  SpUserModeInitialize(
    ULONG LsaVersion,
    PULONG PackageVersion,
    PSECPKG_USER_FUNCTION_TABLE *ppTables,
    PULONG pcTables
  )
{
	return MSV1_0_SpUserModeInitialize(LsaVersion, PackageVersion, ppTables, pcTables);
}


/*
 * MSV1_0 interface
 */

#ifdef EXPORT_MSV_INTERFACE
NTSTATUS NTAPI
  MsvSamValidate(
    PVOID Unknown1,
    PVOID Unknown2,
    PVOID Unknown3,
    PVOID Unknown4,
    PVOID Unknown5,
    PVOID Unknown6,
    PVOID Unknown7,
    PVOID Unknown8,
    PVOID Unknown9,
    PVOID Unknown10,
    PVOID Unknown11,
    PVOID Unknown12,
    PVOID Unknown13
  )
{
	return MSV1_0_SamValidate(Unknown1, Unknown2, Unknown3, Unknown4, Unknown5, Unknown6,
		Unknown7, Unknown8, Unknown9, Unknown10, Unknown11, Unknown12, Unknown13);
}


NTSTATUS NTAPI
  MsvSamLogoff(
    PVOID Unknown1,
    PVOID Unknown2,
    PVOID Unknown3
  )
{
	return MSV1_0_SamLogoff(Unknown1, Unknown2, Unknown3);
}


NTSTATUS NTAPI
  MsvGetLogonAttemptCount(
  )
{
	return MSV1_0_GetLogonAttemptCount();
}


NTSTATUS NTAPI
  Msv1_0ExportSubAuthenticationRoutine(
    PVOID Unknown1,
    PVOID Unknown2,
    PVOID Unknown3,
    PVOID Unknown4,
    PVOID Unknown5,
    PVOID Unknown6,
    PVOID Unknown7,
    PVOID Unknown8,
    PVOID Unknown9,
    PVOID Unknown10
  )
{
	return MSV1_0_ExportSubAuthenticationRoutine(Unknown1, Unknown2, Unknown3, Unknown4,
		Unknown5, Unknown6, Unknown7, Unknown8, Unknown9, Unknown10);
}


NTSTATUS NTAPI
  Msv1_0SubAuthenticationPresent(
    PVOID Unknown1
  )
{
	return MSV1_0_SubAuthenticationPresent(Unknown1);
}
#endif