/*
 * ypauth: YP authentication DLL for Windows 2000
 * Copyright (C) Matthew Chapman 2001-2002
 */

#define SECURITY_WIN32
#include <windows.h>
#include <sspi.h>
#include <ntsecapi.h>
#include <ntsecpkg.h>

#include <rpc/rpc.h>
#include "yp.h"
#include "ufc-crypt.h"

/* Configurables */
#define ADMINISTRATOR		L"Administrator"
#define LOGON_SCRIPT		L"scselogin.bat"
#define LOGON_SERVER		L"SMBDC-SCSE"
#define YP_DOMAIN		"orchestra.cse.unsw.EDU.AU"
#define YP_SERVER_1		"britten." YP_DOMAIN
#define YP_SERVER_2		"maestro." YP_DOMAIN
#define YP_MAP			"passwd.winnt"
#define YP_RID_BASE		1000
#define MSV1_0_DLL_NAME		"MSV1_0_DLL" /* the real mccoy */
#define EXPORT_MSV_INTERFACE

/* Internal defines */
#define CRYPT_LEN		13
#define LOGON_SERVER_LEN	(sizeof(LOGON_SERVER))
#define LOGON_SCRIPT_LEN	(sizeof(LOGON_SCRIPT))

/* Relevant NTSTATUS values from NT DDK */
#define STATUS_SUCCESS			((NTSTATUS)0x00000000L)
#define STATUS_INVALID_PARAMETER        ((NTSTATUS)0xC000000DL)
#define STATUS_NO_LOGON_SERVERS		((NTSTATUS)0xC000005EL)
#define STATUS_LOGON_FAILURE		((NTSTATUS)0xC000006DL)

/* Undocumented LsaApMsInitialize API function */
typedef NTSTATUS (NTAPI *PLSA_AP_MS_INITIALIZE)(PVOID);
#define LSA_AP_NAME_MS_INITIALIZE	"LsaApMsInitialize\0"

/* Prototypes for non-API functions exported from MSV1_0 */
#ifdef EXPORT_MSV_INTERFACE
typedef NTSTATUS (*MSV_SAM_VALIDATE)(PVOID, PVOID, PVOID, PVOID, PVOID, PVOID,
				PVOID, PVOID, PVOID, PVOID, PVOID, PVOID, PVOID);
typedef NTSTATUS (*MSV_SAM_LOGOFF)(PVOID, PVOID, PVOID);
typedef NTSTATUS (*MSV_GET_LOGON_ATTEMPT_COUNT)();
typedef NTSTATUS (*MSV_EXPORT_SUB_AUTHENTICATION_ROUTINE)(PVOID, PVOID, PVOID,
				PVOID, PVOID, PVOID, PVOID, PVOID, PVOID, PVOID);
typedef NTSTATUS (*MSV_SUB_AUTHENTICATION_PRESENT)(PVOID);

#endif

/* Prototypes for NTDLL functions */
VOID NTAPI
  RtlInitUnicodeString(
    PUNICODE_STRING DestinationString,
    PCWSTR SourceString
  );

VOID NTAPI
  RtlRunDecodeUnicodeString(
    UCHAR DecodeSeed,
    PUNICODE_STRING String
  );

LONG NTAPI 
  RtlCompareUnicodeString(
    PUNICODE_STRING String1,
    PUNICODE_STRING String2,
    BOOLEAN CaseInSensitive
  );

NTSTATUS NTAPI
  RtlUnicodeStringToAnsiString(
    PSTRING DestinationString,
    PUNICODE_STRING SourceString,
    BOOLEAN AllocateDestinationString
  );

VOID NTAPI
  RtlFreeAnsiString(
    IN PSTRING AnsiString
  );

/* Local prototypes */
PSID
  DuplicateSid(
    PSID Sid
  );

PSID
  AppendSubAuthority(
    PSID DomainSid,
    DWORD Rid
  );

PUNICODE_STRING
  DuplicateUnicodeString(
    PUNICODE_STRING UnicodeString
  );

void
  CopyUnicodeString(
    PUNICODE_STRING DestString,
    PUNICODE_STRING SourceString
  );

NTSTATUS
  YpLogon(
    char *server,
    PSTRING UserName,
    PSTRING Password,
    PUSHORT Uid
  );

NTSTATUS
  InitializeDomain(
    POLICY_ACCOUNT_DOMAIN_INFO **Info
  );

void
  CreateProfile(
    PLSA_CLIENT_REQUEST ClientRequest,
    PVOID* ProfileBuffer,
    PULONG ProfileBufferSize
  );

char *
  crypt(
    char *key,
    char *salt
  );
